class common(object):
    DISPLAYOFF = 0xAE
    DISPLAYON = 0xAF
    DISPLAYALLON = 0xA5
    DISPLAYALLON_RESUME = 0xA4
    NORMALDISPLAY = 0xA6
    INVERTDISPLAY = 0xA7
    SETREMAP = 0xA0
    SETMULTIPLEX = 0xA8
    SETCONTRAST = 0x81
