# classes to perform actions with frames
from .capture_action_base import CaptureActionBase

# enum with available capture actions
from .capture_actions import CaptureActions
from .generic_action import GenericAction
from .motion_detector import MotionDetector
from .store_frame import StoreFrame
from .video_capture import VideoCapture
