from .componentable import Componentable
from .recreatable import Recreatable
from .stateful import Stateful
from .supports_battery import SupportsBattery
from .supports_miniscreen import SupportsMiniscreen
